/* -----------------------------------------------------------------
The VIVA file system - version 2.0

Implemented by Shankar Pasupathy, July 1996. (shankar@pop.uky.edu)
Modified by Vamsi Krishna, July 1997. (vamsi@cslab.uky.edu)
VIVA was developed by Eric Herrin II and Raphael Finkel.
Department of Computer Science,
University of Kentucky, Lexington 40506.
Distribution site: http://www.cslab.uky.edu/~vamsi/viva

This implementation uses code from the Minix filesystem,
(Copyright: Linus Torvalds).

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation. Read the file COPYRIGHT, included with
this distribution, for further details.
--------------------------------------------------------------------*/

/*  Viva's file handling primitives */
/*  Unlike Ext2 and Minix, we can't use VFS's default lseek function */

/*
#ifdef MODULE
#include <linux/module.h>
#endif
*/

#include <asm/segment.h>
#include <asm/system.h>

#include <linux/sched.h>
#include <linux/viva_fs.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/stat.h>
#include <linux/locks.h>
#include <linux/pagemap.h>

#define	NBUF	32

#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

#include <linux/fs.h>
#include <linux/viva_fs.h>

#define Count_bits_set(byte) (Bits_set[*(unsigned char *)(byte)])
 
unsigned char Bits_set[256] = 
{ 									/* Number of 1s in a byte */
	0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4,
	1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
	1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
	2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
	1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
	2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
	2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
	3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
	1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
	2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
	2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
	3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
	2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
	3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
	3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
	4,5,5,6,5,6,6,7,5,6,6,7,6,7,7,8
};


int Zeros[256] =
{                                   /* Number of zeros in a byte */ 
    8,7,7,6,7,6,6,5,7,6,6,5,6,5,5,4,
	7,6,6,5,6,5,5,4,6,5,5,4,5,4,4,3,
	7,6,6,5,6,5,5,4,6,5,5,4,5,4,4,3,
	6,5,5,4,5,4,4,3,5,4,4,3,4,3,3,2,
	7,6,6,5,6,5,5,4,6,5,5,4,5,4,4,3,
	6,5,5,4,5,4,4,3,5,4,4,3,4,3,3,2,
	6,5,5,4,5,4,4,3,5,4,4,3,4,3,3,2,
	5,4,4,3,4,3,3,2,4,3,3,2,3,2,2,1,
	7,6,6,5,6,5,5,4,6,5,5,4,5,4,4,3,
	6,5,5,4,5,4,4,3,5,4,4,3,4,3,3,2,
	6,5,5,4,5,4,4,3,5,4,4,3,4,3,3,2,
	5,4,4,3,4,3,3,2,4,3,3,2,3,2,2,1,
	6,5,5,4,5,4,4,3,5,4,4,3,4,3,3,2,
	5,4,4,3,4,3,3,2,4,3,3,2,3,2,2,1,
	5,4,4,3,4,3,3,2,4,3,3,2,3,2,2,1,
	4,3,3,2,3,2,2,1,3,2,2,1,2,1,1,0,
}; 

																	
static int viva_file_read(struct inode *, struct file *, char *, int);
static int viva_file_write(struct inode *, struct file *,const char *, int);
static int viva_file_release(struct inode *, struct file *);
static struct buffer_head * viva_new_frag(struct inode * , int);
void viva_free_frag(struct inode *, int , int );

/*
 * We have mostly NULL's here: the current defaults are ok for
 * the viva filesystem.
 */
static struct file_operations viva_file_operations = {
	viva_lseek,				/* Need a special version of lseek */
	viva_file_read,			/* read */
	viva_file_write,		/* write */
	NULL,					/* readdir - bad */
	NULL,					/* select - default */
	NULL,					/* ioctl - default */
	generic_file_mmap,  	/* mmap */
	NULL,					/* no special open is needed */
	viva_file_release,		/* release */
	viva_sync_file			/* fsync */
};

struct inode_operations viva_file_inode_operations = {
	&viva_file_operations,	/* default file operations */
	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	generic_readpage,       /* readpage */
	NULL, 			/* writepage */
	viva_bmap,		/* bmap */
	viva_truncate,		/* truncate */
	NULL			/* permission */ 
};



static int viva_file_read(struct inode * inode, struct file * filp, char * buf, int count)
{
	int read,left,chars;
	int block, blocks, offset;
	int bhrequest, uptodate;
	struct buffer_head ** bhb, ** bhe;
	struct buffer_head * bhreq[NBUF];
	struct buffer_head * buflist[NBUF];
	unsigned int size;
	unsigned int chunk_size=VIVA_BLOCK_SIZE;
	int frag_flag=0;
	int frag_offset=0;
	int blk_count;
	int f_num;
	struct buffer_head *bh;

	if (!inode) {
		printk("viva_file_read: inode = NULL\n");
		return -EINVAL;
	}
	if (!S_ISREG(inode->i_mode)) {
		printk("viva_file_read: mode = %07o\n",inode->i_mode);
		return -EINVAL;
	}
	offset = filp->f_pos;
	size = inode->i_size;
	if (offset > size)
		left = 0;
	else
		left = size - offset;
	if (left > count)
		left = count;
	if (left <= 0)
		return 0;
	read = 0;
	block = offset >> VIVA_BLOCK_SIZE_BITS;
	offset &= VIVA_BLOCK_SIZE-1;
	size = (size + (VIVA_BLOCK_SIZE-1)) >> VIVA_BLOCK_SIZE_BITS;
	blocks = (left + offset + VIVA_BLOCK_SIZE - 1) >> VIVA_BLOCK_SIZE_BITS;
	bhb = bhe = buflist;
	if (filp->f_reada) {
		if(blocks < read_ahead[MAJOR(inode->i_dev)] / (VIVA_BLOCK_SIZE >> 9))
			blocks = read_ahead[MAJOR(inode->i_dev)] / (VIVA_BLOCK_SIZE >> 9);
		if (block + blocks > size)
			blocks = size - block;
	}

/*printk("Vamsi: %d Blocks read for this file, left = %d, frag_blocks = %d \n",blocks,left,inode->u.viva_i.i_fragsInUse);*/

	do 
	{
		bhrequest = 0;
		uptodate = 1;
		blk_count = 0;
		frag_flag = 0;
		chunk_size = VIVA_BLOCK_SIZE;
		while (blocks) {
			--blocks;
			if(inode->u.viva_i.i_fragsInUse 
				&& (block == inode->u.viva_i.i_blocksInUse))
			{
				frag_flag = 1;
				break;
			}
			blk_count++;
			*bhb = viva_getblk(inode, block++, 0);
			if (*bhb && !buffer_uptodate(*bhb)) {
				uptodate = 0;
				bhreq[bhrequest++] = *bhb;
			}

			if (++bhb == &buflist[NBUF])
				bhb = buflist;

			/* If the block we have on hand is uptodate, go ahead
			   and complete processing. */
			if (uptodate)
				break;
			if (bhb == bhe)
				break;
		}

/*		printk("Vamsi: Reached till here 1 ....%d \n", left);*/
		/* Now request them all */
		if (bhrequest)
			ll_rw_block(READ, bhrequest, bhreq);
		if(blk_count || !frag_flag)
		{
			do 
			{ /* Finish off all I/O that has actually completed */
				if (*bhe) {
					wait_on_buffer(*bhe);
					if (!buffer_uptodate(*bhe)){	/* read error? */
							brelse(*bhe);
						if (++bhe == &buflist[NBUF])
						  bhe = buflist;
						left = 0;
						break;
					}
				}

				if (left < chunk_size - offset)
					chars = left;
				else
					chars = chunk_size - offset;
				filp->f_pos += chars;
				left -= chars;
				read += chars;
				if (*bhe) {
					memcpy_tofs(buf,offset+(*bhe)->b_data,chars);
					brelse(*bhe);
					buf += chars;
				} else {
					while (chars-->0)
						put_fs_byte(0,buf++);
				}
				offset = 0;
				if (++bhe == &buflist[NBUF])
					bhe = buflist;
			} while (bhe != bhb && (!*bhe || !buffer_locked(*bhe)));
		}
		if(frag_flag && inode->u.viva_i.i_fragsInUse)	
		{
			chunk_size = VIVA_FRAG_SIZE;
			/* Vamsi: Read all the fragments here ... */
/*			printk("Vamsi: Trying to read the Fragment part of the file \n");*/

			bh = bread(inode->i_dev, inode->u.viva_i.i_frag_block,
														VIVA_BLOCK_SIZE);
			for(f_num=0; f_num<inode->u.viva_i.i_fragsInUse && left>0; f_num++)	
			{
				if(offset >= chunk_size)
				{
					offset -= chunk_size;
					continue;
				}
				frag_offset =(inode->u.viva_i.i_frag_no[f_num]*VIVA_FRAG_SIZE);
				if (left < chunk_size - offset)
					chars = left;
				else
					chars = chunk_size - offset;
				filp->f_pos += chars;
				left -= chars;
				read += chars;
				if (bh) {
/*				printk("Vamsi: Frag_block read = %d (%d), left = %d, pos= %d \n",inode->u.viva_i.i_frag_block, inode->u.viva_i.i_frag_no[f_num],left,filp->f_pos);*/
					memcpy_tofs(buf,offset+frag_offset+bh->b_data,chars);
/*					brelse(bh);*/
					buf += chars;
				} else {
					while (chars-->0)
						put_fs_byte(0,buf++);
				}
				offset = 0;
			}
		}

	} while (left > 0);

/* Release the read-ahead blocks */
	while (bhe != bhb) {
		brelse(*bhe);
		if (++bhe == &buflist[NBUF])
			bhe = buflist;
	}
	if (!read)
		return -EIO;
	filp->f_reada = 1;
	if (!IS_RDONLY(inode))
		inode->i_atime = CURRENT_TIME;
	return read;
}


static int viva_file_write(struct inode * inode, struct file * filp, const char * buf, int count)
{
	off_t pos;
	int written,c=0;
	struct buffer_head * bh;
	char *ptr=NULL;
	int offset=0;

	if (!inode) {
		printk("viva_file_write: inode = NULL\n");
		return -EINVAL;
	}
	if (!S_ISREG(inode->i_mode)) {
		printk("viva_file_write: mode = %07o\n",inode->i_mode);
		return -EINVAL;
	}
/*	down(&inode->i_sem);*/
	if (filp->f_flags & O_APPEND)
		pos = inode->i_size;
	else
		pos = filp->f_pos;
	written = 0;
	while (written < count) 
	{
		bh = viva_getblk(inode,pos/VIVA_BLOCK_SIZE,1);
/*		printk("Vamsi: Allocated Block %d for the file at %d\n",bh->b_blocknr,pos); */
		if (!bh) 
		{
			if (!written)
				written = -ENOSPC;
			break;
		}
		c = VIVA_BLOCK_SIZE - (pos % VIVA_BLOCK_SIZE);
		if (c > count-written)
			c = count-written;
		if (c != VIVA_BLOCK_SIZE && !buffer_uptodate(bh)) {
			ll_rw_block(READ, 1, &bh);
			wait_on_buffer(bh);
			if (!buffer_uptodate(bh)) {
				brelse(bh);
				if (!written)
					written = -EIO;
				break;
			}
		}
		ptr = (pos % VIVA_BLOCK_SIZE) + offset + bh->b_data;
		memcpy_fromfs(ptr,buf,c);
		update_vm_cache(inode,pos,ptr,c);
		mark_buffer_uptodate(bh,1);
		mark_buffer_dirty(bh, 0);
		brelse(bh);
		buf += c;
		pos += c;
		written += c;
		offset = 0;
	}
	if (pos > inode->i_size)
		inode->i_size = pos;
/*	up(&inode->i_sem);*/
	inode->i_mtime = inode->i_ctime = CURRENT_TIME;
	filp->f_pos = pos;
	inode->i_dirt = 1;
	return written;
}


static int viva_file_release(struct inode *inode, struct file *filp)
{
	struct buffer_head *bh=NULL;
	struct buffer_head *frag_bh=NULL;
	unsigned long prev_block_addr;
	unsigned long last_block;
	int left, size=VIVA_FRAG_SIZE;
	int count, num_frags;

	if(filp->f_mode & 2)
	{
		left = inode->i_size%VIVA_BLOCK_SIZE;
		last_block = inode->i_size/VIVA_BLOCK_SIZE;
/*		printk("Vamsi: Closing a file which was opened for writing \n",);*/
		if( left && (left <= (VIVA_FRAG_SIZE*(VIVA_FRAG_PER_BLOCK-1)))
				 && (inode->i_sb->u.viva_sb.s_frag_blocks 
						< inode->i_sb->u.viva_sb.s_max_frag_blocks) )
		{
			bh = viva_getblk(inode, last_block, 0);
			/* Vamsi: Second parameter is not really required */
			if(!bh)
			{
				printk("file_release: Unable to read the last block %ld of this file \n",inode->i_size/VIVA_BLOCK_SIZE);	
				brelse(bh);
				return 1;
			}
			prev_block_addr = viva_release_smap(inode->u.viva_i.i_cached_smap,
													last_block);
			if(prev_block_addr < 0)
			{
				printk("file_release: Unable to fragment the last block \n");
				return(-1);
			}
			num_frags = (left/VIVA_FRAG_SIZE);
			if(left%VIVA_FRAG_SIZE)
				num_frags++;
			frag_bh = viva_new_frag(inode,num_frags);
/*			printk("Vamsi: Block %ld (%ld) is replaced by ",bh->b_blocknr, prev_block_addr); */
			for(count=0; count<num_frags; count++)
			{
				if(count == left/VIVA_FRAG_SIZE)
					size = left%VIVA_FRAG_SIZE;
				bcopy( bh->b_data+(count*VIVA_FRAG_SIZE), frag_bh->b_data +
					(inode->u.viva_i.i_frag_no[count] * VIVA_FRAG_SIZE), size );
/*				printk("%ld (%d %d) ",frag_bh->b_blocknr, inode->u.viva_i.i_frag_no[count],size);*/
				mark_buffer_uptodate(frag_bh,1);
				mark_buffer_dirty(frag_bh,1);
			}

/*			printk(" %d frag_blocks \n",inode->u.viva_i.i_fragsInUse);*/
			viva_free_block(inode->i_sb, bh->b_blocknr);
			inode->u.viva_i.i_lastblock = prev_block_addr;
			inode->i_ctime = CURRENT_TIME;
			inode->u.viva_i.i_blocksInUse = inode->i_size/VIVA_BLOCK_SIZE;
			inode->i_dirt = 1;
			brelse(bh);
			brelse(frag_bh);
		}
	}
	return 1;
}


static struct buffer_head * viva_new_frag(struct inode * inode, int num_frags)
{
	unsigned int new_block_addr;
	unsigned int goal, count, frags_used;
	struct buffer_head * result;
	struct buffer_head * frag_bh;
	unsigned int fa_map_slot, fa_map_byte, fa_map_bit;
	int frag_num, frag_block_no=0;
	int expand_retval, temp_bit, block_nr;
	unsigned char temp_byte;
	

	if(num_frags == inode->u.viva_i.i_fragsInUse)
	{
		if (!(result = getblk(inode->i_sb->s_dev, inode->u.viva_i.i_frag_block, 
															VIVA_BLOCK_SIZE)))
		{
			printk("viva_new_frag: Unable to get the Fragment block\n");
			return NULL;
		}
		/* Frag_block already exists */
		inode->i_dirt = 1;
		return result;
	}
	if(inode->u.viva_i.i_fragsInUse)
	{
		frags_used = inode->u.viva_i.i_fragsInUse;
		for(count=0; count < frags_used; count++)
		{
			viva_free_frag(inode, inode->u.viva_i.i_frag_block, 
										inode->u.viva_i.i_frag_no[count]);
			inode->u.viva_i.i_frag_no[count] = 0;
		}
		inode->u.viva_i.i_frag_block = 0;
		inode->u.viva_i.i_fragsInUse = 0;
	}
	/* Number of more Fragments required */
	while(!inode->u.viva_i.i_fragsInUse 
			&& (frag_block_no<inode->i_sb->u.viva_sb.s_max_frag_blocks))
	{
		fa_map_slot = (frag_block_no*VIVA_FRAG_PER_BLOCK)
								/(VIVA_BLOCK_SIZE << 3);
		fa_map_byte = ((frag_block_no*VIVA_FRAG_PER_BLOCK)/8) 
								-(fa_map_slot*VIVA_BLOCK_SIZE);
		fa_map_bit = (frag_block_no*VIVA_FRAG_PER_BLOCK)%8; 

		frag_bh = inode->i_sb->u.viva_sb.s_famap[fa_map_slot];		
		if (!frag_bh) 
		{ 
			printk("viva_new_frag: nonexistent bitmap buffer\n");
			return NULL;
		}
		temp_byte = *(frag_bh->b_data+fa_map_byte) ;
		if(fa_map_bit)
			temp_byte = temp_byte | 0x0f;
		else
			temp_byte = temp_byte | 0xf0;

		if(Zeros[temp_byte] >= num_frags)
		{
			if(Zeros[temp_byte] == 4) /* If all the 4 fragments are unused */
				inode->i_sb->u.viva_sb.s_frag_blocks++;
			temp_bit=find_first_zero_bit(&temp_byte,8);
/*			printk("new_frag: In a block, fragments ");*/
			for(frag_num=0; frag_num < num_frags; frag_num++)
			{
				if(temp_bit > VIVA_FRAG_PER_BLOCK+fa_map_bit)
					printk("viva_new_frag: Error in Frag_alloc bitmap \n");
/*				printk("%d ",temp_bit%VIVA_FRAG_PER_BLOCK);*/
				set_bit(temp_bit,frag_bh->b_data+fa_map_byte);
				inode->u.viva_i.i_frag_no[frag_num] = 
										temp_bit%VIVA_FRAG_PER_BLOCK;
				inode->u.viva_i.i_fragsInUse++;

				temp_bit = find_next_zero_bit(&temp_byte,8,temp_bit+1);
			}
/*			printk(" are allocated \n");*/
			mark_buffer_uptodate(frag_bh,1);
			mark_buffer_dirty(frag_bh,1);
		}
		else
			frag_block_no++;
		
	} 
	block_nr = viva_bmap(inode->i_sb->u.viva_sb.s_cached_fsmap, frag_block_no);
	if(block_nr)
	{
		if (!(result = getblk(inode->i_sb->s_dev, block_nr, VIVA_BLOCK_SIZE)))
		{
			printk("viva_new_frag: Unable to get a Fragment block\n");
			return NULL;
		}
		/* Frag_block already exists */
/*		printk("Block %d (%d) is allocated for the above fragments \n", block_nr, frag_block_no);*/
		inode->u.viva_i.i_frag_block = block_nr;
		inode->i_dirt = 1;
		return result;
	}

	goal = inode->i_sb->u.viva_sb.s_last_frag_block;
	new_block_addr = viva_new_block(inode->i_sb,goal,8);
	if (new_block_addr < 0)
	{
		/* No blocks available or some error */
		printk("viva_new_frag: No more blocks available for fragments \n");
		return NULL;
	}
	if(new_block_addr > inode->i_sb->u.viva_sb.s_last_frag_block)
	{
		inode->i_sb->u.viva_sb.s_last_frag_block = new_block_addr;
	}

	inode->i_sb->u.viva_sb.s_frag_blocks++;
	inode->u.viva_i.i_frag_block = new_block_addr;
/*	printk("New Block %d (%d) is allocated for the above fragments \n", new_block_addr, frag_block_no);*/

	if ((expand_retval = viva_expand_smap(inode->i_sb->u.viva_sb.s_cached_fsmap,
										new_block_addr, VIVA_FS_MAP_SIZE)) < 0) 
	{
		printk("VIVA: No space left to expand Fragment sparse bitmap\n");
		viva_free_block(inode->i_sb,new_block_addr);
		return NULL;
	}
	if (!(result = getblk(inode->i_sb->s_dev,new_block_addr,VIVA_BLOCK_SIZE)))
	{
		printk("VIVA: new_frag: cannot get a Fragment block\n");
		return NULL;
	}

	inode->i_dirt = 1;
	return result;

}


void viva_free_frag(struct inode *inode, int frag_block, int frag_num)
{
	struct buffer_head * frag_bh;
	unsigned int fa_map_slot, fa_map_byte, fa_map_bit;
	int byte, frag_block_no=0;
	char temp_byte[1];

	unsigned long last_valid_addr,first_valid_addr, block_addr;
	struct viva_smap_entry *entry;
	unsigned char mask;
	
	entry = (struct viva_smap_entry *)(inode->i_sb->u.viva_sb.s_cached_fsmap);
	last_valid_addr = (unsigned long)(entry) + VIVA_FS_MAP_SIZE - 1;
	first_valid_addr = (unsigned long)(entry);
	frag_block_no = 0;
	while(entry->start_addr)
	{
		if( (entry->start_addr <= frag_block) 
			&& (frag_block <= entry->start_addr+(entry->length*8)) ) 
		{
			if(entry->start_addr == frag_block)
				goto found_block;
			frag_block_no++; /* because the first bit is implicitly set */
			byte = 0;
			block_addr = entry->start_addr;
			while(byte < entry->length)
			{
				block_addr += 8;
				if(frag_block <= block_addr)
					goto block_in_this_byte;
				frag_block_no += Count_bits_set(entry->bitmap + byte);
				byte++;
			}
			/* block requested in not available */
			printk("Viva_free_frag: Error in removing a fragment \n");
			return;
		}

		frag_block_no++; /* because the first bit is implicitly set */
		byte = 0;
		while(byte < entry->length)
		{
			frag_block_no += Count_bits_set(entry->bitmap + byte);
			byte++;
		}
		entry = (struct viva_smap_entry *) (entry->bitmap + entry->length);
	}
	/* block requested in not available */
	printk("Viva_free_frag: Unable to free fragments in %d block\n",frag_block);
	return;

block_in_this_byte:
	memcpy(temp_byte,(entry->bitmap+byte),1);
	mask = 255 << (7 - (block_addr-frag_block));
	mask = mask ^ 0xff;
	*temp_byte &= mask;
	frag_block_no += Count_bits_set(temp_byte);

found_block:
/*	printk("free_frag: frag_block_no = %d (%d %d) (%d %d)\n", frag_block_no,frag_block, frag_num, *temp_byte, mask);*/
	fa_map_slot = (frag_block_no*VIVA_FRAG_PER_BLOCK) /(VIVA_BLOCK_SIZE << 3);
	fa_map_byte = ((frag_block_no*VIVA_FRAG_PER_BLOCK)/8) 
								-(fa_map_slot*VIVA_BLOCK_SIZE);
	fa_map_bit = (frag_block_no*VIVA_FRAG_PER_BLOCK)%8; 

	frag_bh = inode->i_sb->u.viva_sb.s_famap[fa_map_slot];		
	if (!frag_bh) 
	{ 
		printk("viva_free_frag: nonexistent bitmap buffer\n");
		return;
	}
	if (!clear_bit(frag_num+fa_map_bit,frag_bh->b_data+fa_map_byte)) 
	{ 
		printk("viva_free_frag: bit %d already cleared in block %d \n",
														frag_num,frag_block);
	}
	mark_buffer_uptodate(frag_bh,1);
	mark_buffer_dirty(frag_bh,1);
	*temp_byte = *(frag_bh->b_data+fa_map_byte);
	if(fa_map_bit)
		*temp_byte = *temp_byte & 0xf0;
	else
		*temp_byte = *temp_byte & 0x0f;
	if( !(*temp_byte))
		if(inode->i_sb->u.viva_sb.s_frag_blocks)
			inode->i_sb->u.viva_sb.s_frag_blocks--;

	inode->u.viva_i.i_fragsInUse--;
	inode->i_dirt = 1;
	return;
}	
