/* -----------------------------------------------------------------
The VIVA file system - version 2.0

Implemented by Shankar Pasupathy, July 1996. (shankar@pop.uky.edu)
Modified by Vamsi Krishna, July 1997. (vamsi@cslab.uky.edu)
VIVA was developed by Eric Herrin II and Raphael Finkel.
Department of Computer Science,
University of Kentucky, Lexington 40506.
Distribution site: http://www.cslab.uky.edu/~vamsi/viva

This implementation uses code from the Minix filesystem,
(Copyright: Linus Torvalds).

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation. Read the file COPYRIGHT, included with
this distribution, for further details.
--------------------------------------------------------------------*/
/* File synchronization primitives - essentially write data in the buffer 
cache to disk */

/*
#ifdef MODULE
#include <linux/module.h>
#endif
*/

#include <asm/segment.h>
#include <asm/system.h>

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/fcntl.h>
#include <linux/locks.h>

#include <linux/fs.h>
#include <linux/viva_fs.h>

#define blocksize VIVA_BLOCK_SIZE

static int sync_block (struct inode * inode, unsigned long * block, int wait)
{
	struct buffer_head * bh;
	unsigned long tmp;
	
	if (!*block)
		return 0;
	tmp = *block;
	bh = get_hash_table(inode->i_dev, *block, blocksize);
	if (!bh)
		return 0;
	if (*block != tmp) {
		brelse (bh);
		return 1;
	}
	if (wait && buffer_req(bh) && !buffer_uptodate(bh)) {
		brelse(bh);
		return -1;
	}
	if (wait || !buffer_uptodate(bh) || !buffer_dirty(bh))
	{
		brelse(bh);
		return 0;
	}
	ll_rw_block(WRITE, 1, &bh);
	bh->b_count--;
	return 0;
}



/*--------------------------------------------------
sync_direct is a misnomer. This function actually flushes
direct as well as indirect blocks in Viva.
--------------------------------------------------- */

static int sync_direct(struct inode *inode, int wait)
{
	int i=0;
	unsigned long tmp;
	int rc, err = 0;
	int num_blocks;

	num_blocks = inode->u.viva_i.i_blocksInUse;
	while(num_blocks)
	{
		tmp = viva_bmap(inode->u.viva_i.i_cached_smap,i); 
		if(tmp > 0)
			rc = sync_block (inode, &tmp, wait);
		else
			printk("viva_fsync: From here \n");
		if (rc > 0)
			break;
		if (rc)
			err = rc;
		i++;
		num_blocks--;
	}
	return err;
}

int viva_sync_file(struct inode * inode, struct file * file)
{
	int wait, err = 0;
	
	if (!(S_ISREG(inode->i_mode) || S_ISDIR(inode->i_mode) ||
	     S_ISLNK(inode->i_mode)))
		return -EINVAL;

	for (wait=0; wait<=1; wait++)
	{
		err |= sync_direct(inode, wait);
	}
	err |= viva_sync_inode (inode);
	return (err < 0) ? -EIO : 0;
}
