/* -----------------------------------------------------------------
The VIVA file system - version 2.0

Implemented by Shankar Pasupathy, July 1996. (shankar@pop.uky.edu)
Modified by Vamsi Krishna, July 1997. (vamsi@cslab.uky.edu)
VIVA was developed by Eric Herrin II and Raphael Finkel.
Department of Computer Science,
University of Kentucky, Lexington 40506.
Distribution site: http://www.cslab.uky.edu/~vamsi/viva

This implementation uses code from the Minix filesystem,
(Copyright: Linus Torvalds).

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation. Read the file COPYRIGHT, included with
this distribution, for further details.
--------------------------------------------------------------------*/

/* Really difficult code to implement. Had lot of problems with this
code even though I followed minix/truncate.c closely. Be very careful
when you modify any of these functions */
/* The function trunc_direct is a misnomer since it truncates both 
direct and indirect blocks */

/*
#ifdef MODULE
#include <linux/module.h>
#endif
*/

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/viva_fs.h>
#include <linux/stat.h>
#include <linux/fcntl.h>

/* Linus Torvalds' comments:
 *
 * Truncate has the most races in the whole filesystem: coding it is
 * a pain in the a**. Especially as I don't do any locking...
 *
 * The code may look a bit weird, but that's just because I've tried to
 * handle things like file-size changes in a somewhat graceful manner.
 * Anyway, truncating a file at the same time somebody else writes to it
 * is likely to result in pretty weird behaviour...
 *
 * The new code handles normal truncates (size = 0) as well as the more
 * general case (size = XXX). I hope.
 */

/* -------------------------------------------------------------
trunc_direct: Poor name for this function. In VIVA, it actually
truncates direct as well as indirect blocks. 
-------------------------------------------------------------- */
static int viva_trunc_direct(struct inode * inode)
{
	unsigned int block_addr;
	struct buffer_head * bh;
	int i, tmp;
	int retry = 0;
	int count, num_frags;
	i = 0;
	while ((block_addr = viva_bmap(inode->u.viva_i.i_cached_smap,i)) > 0) {
		tmp = block_addr;
		bh = get_hash_table(inode->i_dev,tmp,VIVA_BLOCK_SIZE);
		if ((bh && bh->b_count != 1) || tmp != block_addr) {
			retry = 1;
			brelse(bh);
			continue;
		}
/*		printk("Vamsi: Truncating the file  ....\n");*/
		if (tmp == inode->u.viva_i.i_lastblock) {
			/* last block is being freed */
			/* clear all sparse bitmap entries */
			memset(inode->u.viva_i.i_cached_smap,0,VIVA_SMAP_SIZE);	
			/* set the lastblock number to 0 */
			inode->u.viva_i.i_lastblock = 0;
		}
		viva_free_block(inode->i_sb,tmp);
/*		inode->u.viva_i.i_blocksInUse--;*/
		/* The memory version of the inode has been dirtied. So write it's
		contents to disk. setting the variable i_dirt does that. */
		i++;
		inode->i_dirt = 1;
		brelse(bh);
	}
	if (inode->u.viva_i.i_fragsInUse)
	{
		num_frags = inode->u.viva_i.i_fragsInUse;
		for(count=0; count < num_frags; count++)
		{
			viva_free_frag(inode, inode->u.viva_i.i_frag_block, 
										inode->u.viva_i.i_frag_no[count]);
			inode->u.viva_i.i_frag_no[count] = 0;
		}
		inode->u.viva_i.i_frag_block = 0;
	}
	inode->u.viva_i.i_blocksInUse = 0;
	inode->u.viva_i.i_fragsInUse = 0;
	return retry;
}

void viva_truncate(struct inode * inode)
{
	int retry;
	
	if (!(S_ISREG(inode->i_mode) || S_ISDIR(inode->i_mode) ||
	     S_ISLNK(inode->i_mode)))
		return;
	while (1) {
		retry = viva_trunc_direct(inode);
		if (!retry)
			break;
		current->counter = 0;
		schedule();
	}
	inode->i_mtime = inode->i_ctime = CURRENT_TIME;
	inode->i_dirt = 1;
}
